/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, 2013, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 * Author: Eitan Marder-Eppstein
 *         David V. Lu!!
 *********************************************************************/
#include <iri_opendrive_global_planner/opendrive_planner.h>
#include <pluginlib/class_list_macros.h>
#include <costmap_2d/cost_values.h>
#include <costmap_2d/costmap_2d.h>
#include <tf2/utils.h>
#include <limits>

#include "exceptions.h"

//register this planner as a BaseOpendriveGlobalPlanner plugin
PLUGINLIB_EXPORT_CLASS(iri_opendrive_global_planner::OpendriveGlobalPlanner, nav_core::BaseGlobalPlanner)

namespace iri_opendrive_global_planner {

OpendriveGlobalPlanner::OpendriveGlobalPlanner() :
  costmap_(NULL), initialized_(false), tf2_listener(tf2_buffer) {
  this->start_angle_tol=3.14159;
  this->start_dist_tol=4.0;
  this->end_angle_tol=3.14159;
  this->end_dist_tol=4.0;
  this->multi_hyp=false;
}

OpendriveGlobalPlanner::OpendriveGlobalPlanner(std::string name, costmap_2d::Costmap2D* costmap, std::string frame_id) :
  OpendriveGlobalPlanner() 
{
  //initialize the planner
  initialize(name, costmap, frame_id);
}

OpendriveGlobalPlanner::~OpendriveGlobalPlanner() 
{
  if (dsrv_)
    delete dsrv_;
}

void OpendriveGlobalPlanner::initialize(std::string name, costmap_2d::Costmap2DROS* costmap_ros) 
{
  initialize(name, costmap_ros->getCostmap(), costmap_ros->getGlobalFrameID());
}

void OpendriveGlobalPlanner::initialize(std::string name, costmap_2d::Costmap2D* costmap, std::string frame_id) 
{
  std::vector<TOSMPathData> osm_paths;

  if (!initialized_) 
  {
    this->private_nh=ros::NodeHandle("~/" + name);
    costmap_ = costmap;
    frame_id_ = frame_id;

    try{
      std::string opendrive_file;
      int cost_type=COST_DIST;
      int utm_zone;
      bool lane_mode=false;

      if(this->private_nh.getParam("utm_zone", utm_zone))
      {
        //ROS_INFO("utm_zone=%d",utm_zone);
        this->roadmap.set_utm_zone(utm_zone);
      }
      else
        ROS_ERROR("OpendriveGlobalPlanner: utm zone not defined");

      if(!this->private_nh.getParam("lane_mode",lane_mode))
        ROS_ERROR("OpendriveGlobalPlanner: Desired planning mode not defined");
      if(lane_mode)
        this->roadmap.set_lane_mode();
      else
        this->roadmap.set_segment_mode();
      if(this->private_nh.getParam("opendrive_file", opendrive_file))
      {
        if(opendrive_file!="")
        {
          try{
            if(opendrive_file.find("xodr")!=std::string::npos)
              this->roadmap.load_opendrive(opendrive_file);
            else if(opendrive_file.find("osm")!=std::string::npos)
            {
              if(this->load_osm_paths(osm_paths))
                this->roadmap.load_osm(opendrive_file,osm_paths);
              else
                ROS_ERROR("Invalid OSM categories information");
            }
            else
              ROS_ERROR("Unknown file type");
            this->create_opendrive_map(opendrive_file,this->roadmap.get_resolution(), utm_zone);
          }catch(CException &e){
            ROS_ERROR_STREAM(e.what());
          }
        }
      }
      else
        ROS_ERROR("OpendriveGlobalPlanner: Opendrive map file not defined");
      if(!this->private_nh.getParam("opendrive_frame", this->opendrive_frame_id_))
        ROS_ERROR("OpendriveGlobalPlanner: Opendrive frame ID not defined");
      if(!this->private_nh.getParam("start_angle_tol", this->start_angle_tol))
        ROS_ERROR("OpendriveGlobalPlanner: Angle tolerance for the start position not defined");
      if(!this->private_nh.getParam("start_dist_tol", this->start_dist_tol))
        ROS_ERROR("OpendriveGlobalPlanner: Distance tolerance for the start position not defined");
      if(!this->private_nh.getParam("end_angle_tol", this->end_angle_tol))
        ROS_ERROR("OpendriveGlobalPlanner: Angle tolerance for the end position not defined");
      if(!this->private_nh.getParam("end_dist_tol", this->end_dist_tol))
        ROS_ERROR("OpendriveGlobalPlanner: Distance tolerance for the end position not defined");
      if(!this->private_nh.getParam("multi_hyp", this->multi_hyp))
        ROS_ERROR("OpendriveGlobalPlanner: Multihypotheis feature not defined");
      if(!this->private_nh.getParam("cost_type",cost_type))
        ROS_ERROR("OpendriveGlobalPlanner: Desired cost type not defined");
      this->roadmap.set_cost_type((cost_t)cost_type);
    }catch(CException &e){
      ROS_ERROR_STREAM(e.what());
    }

    plan_pub_ = this->private_nh.advertise<nav_msgs::Path>("plan", 1);
    make_plan_srv_ = this->private_nh.advertiseService("make_plan", &OpendriveGlobalPlanner::makePlanService, this);

    this->opendrive_map_pub_ = this->private_nh.advertise<nav_msgs::OccupancyGrid>("opendrive_map", 1,boost::bind(&OpendriveGlobalPlanner::map_connect_callback, this, _1));

    this->opendrive_map_service=this->private_nh.advertiseService("get_opendrive_map", &OpendriveGlobalPlanner::get_opendrive_map,this);

    this->opendrive_nodes_service=this->private_nh.advertiseService("get_opendrive_nodes", &OpendriveGlobalPlanner::get_opendrive_nodes,this);

    dsrv_ = new dynamic_reconfigure::Server<iri_opendrive_global_planner::OpendriveGlobalPlannerConfig>(ros::NodeHandle("~/" + name));
    dynamic_reconfigure::Server<iri_opendrive_global_planner::OpendriveGlobalPlannerConfig>::CallbackType cb = boost::bind(&OpendriveGlobalPlanner::reconfigureCB, this, _1, _2);
    dsrv_->setCallback(cb);

    initialized_ = true;
  }
  else
    ROS_WARN("OpendriveGlobalPlanner: This planner has already been initialized, you can't call it twice, doing nothing");
}

void OpendriveGlobalPlanner::map_connect_callback(const ros::SingleSubscriberPublisher &subs)
{
  this->path_access.enter();
  subs.publish(this->full_path_);
  this->path_access.exit();
}

void OpendriveGlobalPlanner::create_opendrive_map(std::string &filename,double resolution, int utm_zone)
{
  std::vector<double> x,partial_x,y,partial_y,heading;
  double max_x=std::numeric_limits<double>::min(),min_x=std::numeric_limits<double>::max(),max_y=std::numeric_limits<double>::min(),min_y=std::numeric_limits<double>::max();
  std::vector<TOSMPathData> osm_paths;
  CRoadMap road;

  this->path_access.enter();
  road.set_utm_zone(utm_zone);
  if(filename.find("xodr")!=std::string::npos)
    road.load_opendrive(filename);
  else if(filename.find("osm")!=std::string::npos)
  {
    if(this->load_osm_paths(osm_paths))
      road.load_osm(filename,osm_paths);
    else
      return;
  }
  else
    return;  
  road.set_resolution(resolution);

  this->full_path_.header.frame_id = this->opendrive_frame_id_;
  this->full_path_.header.stamp = ros::Time::now();
  this->full_path_.info.map_load_time = ros::Time::now();
  this->full_path_.info.resolution = resolution;
  road.get_lane_geometry(x,y,heading);
  for(unsigned int i=0;i<x.size();i++)
  {
    if(x[i]>max_x)
      max_x=x[i];
    else if(x[i]<min_x)
      min_x=x[i];
    if(y[i]>max_y)
      max_y=y[i];
    else if(y[i]<min_y)
      min_y=y[i];
  }
  this->full_path_.info.width=((max_x-min_x)/resolution)+1;
  this->full_path_.info.height=((max_y-min_y)/resolution)+1;
  this->full_path_.info.origin.position.x=min_x;
  this->full_path_.info.origin.position.y=min_y;
  this->full_path_.info.origin.position.z=0.0;
  this->full_path_.info.origin.orientation.x=0.0;
  this->full_path_.info.origin.orientation.y=0.0;
  this->full_path_.info.origin.orientation.z=0.0;
  this->full_path_.info.origin.orientation.w=1.0;
  this->full_path_.data.resize((this->full_path_.info.width)*(this->full_path_.info.height),255);
  for(unsigned int i=0;i<x.size();i++)
    this->full_path_.data[this->full_path_.info.width*((unsigned int)((y[i]-min_y)/resolution))+((unsigned int)((x[i]-min_x)/resolution))]=0;
  this->path_access.exit();
}

bool OpendriveGlobalPlanner::get_opendrive_map(iri_nav_module::get_opendrive_map::Request  &req,iri_nav_module::get_opendrive_map::Response &res)
{
  this->path_access.enter();
  res.opendrive_map=this->full_path_;
  this->path_access.exit();

  return true;
}

bool OpendriveGlobalPlanner::get_opendrive_nodes(iri_nav_module::get_opendrive_nodes::Request  &req,iri_nav_module::get_opendrive_nodes::Response &res)
{
  res.opendrive_nodes.header.frame_id=this->opendrive_frame_id_;
  res.opendrive_nodes.header.stamp=this->best_path_stamp;
  res.opendrive_nodes.nodes.resize(this->best_path.size());
  for(unsigned int i=0;i<this->best_path.size();i++)
    res.opendrive_nodes.nodes[i]=this->best_path[i].segment_id;

  return true;
}

void OpendriveGlobalPlanner::reconfigureCB(iri_opendrive_global_planner::OpendriveGlobalPlannerConfig& new_config, uint32_t level) 
{
  std::vector<TOSMPathData> osm_paths;

  try{
    boost::mutex::scoped_lock lock(mutex_);
    this->roadmap.set_resolution(new_config.resolution);
    this->opendrive_frame_id_=new_config.opendrive_frame;
    this->start_angle_tol=new_config.start_angle_tol;
    this->start_dist_tol=new_config.start_dist_tol;
    this->end_angle_tol=new_config.end_angle_tol;
    this->end_dist_tol=new_config.end_dist_tol;
    this->multi_hyp=new_config.multi_hyp;
    if(new_config.lane_mode)
      this->roadmap.set_lane_mode();
    else
      this->roadmap.set_segment_mode();
    if(new_config.opendrive_file!=this->config.opendrive_file || new_config.lane_mode!=this->config.lane_mode)
    {
      if(new_config.opendrive_file.find("xodr")!=std::string::npos)
        roadmap.load_opendrive(new_config.opendrive_file);
      else if(new_config.opendrive_file.find("osm")!=std::string::npos)
      {
        if(this->load_osm_paths(osm_paths))
          roadmap.load_osm(new_config.opendrive_file,osm_paths);
        else
          ROS_ERROR("Invalid OSM categories information");
      }
      else
        ROS_ERROR("Unknown file type");      
      this->create_opendrive_map(new_config.opendrive_file,new_config.resolution, new_config.utm_zone);
    } 
    this->roadmap.set_cost_type((cost_t)new_config.cost_type);
    this->config=new_config;
  }catch(CException &e){
    ROS_ERROR_STREAM(e.what());
  }
}

void OpendriveGlobalPlanner::clearRobotCell(const geometry_msgs::PoseStamped& global_pose, unsigned int mx, unsigned int my) 
{
  if (!initialized_) 
  {
    ROS_ERROR("OpendriveGlobalPlanner: this planner has not been initialized yet, but it is being used, please call initialize() before use");
    return;
  }

  //set the associated costs in the cost map to be free
  costmap_->setCost(mx, my, costmap_2d::FREE_SPACE);
}

bool OpendriveGlobalPlanner::makePlanService(nav_msgs::GetPlan::Request& req, nav_msgs::GetPlan::Response& resp) 
{
  makePlan(req.start, req.goal, resp.plan.poses);

  resp.plan.header.stamp = ros::Time::now();
  resp.plan.header.frame_id = frame_id_;

  return true;
}

void OpendriveGlobalPlanner::mapToWorld(double mx, double my, double& wx, double& wy) 
{
  wx = costmap_->getOriginX() + mx * costmap_->getResolution();
  wy = costmap_->getOriginY() + my * costmap_->getResolution();
}

bool OpendriveGlobalPlanner::worldToMap(double wx, double wy, double& mx, double& my) 
{
  double origin_x = costmap_->getOriginX(), origin_y = costmap_->getOriginY();
  double resolution = costmap_->getResolution();

  if (wx < origin_x || wy < origin_y)
    return false;

  mx = (wx - origin_x) / resolution;
  my = (wy - origin_y) / resolution;

  if (mx < costmap_->getSizeInCellsX() && my < costmap_->getSizeInCellsY())
    return true;

  return false;
}

bool OpendriveGlobalPlanner::makePlan(const geometry_msgs::PoseStamped& start, const geometry_msgs::PoseStamped& goal,std::vector<geometry_msgs::PoseStamped>& plan) 
{
  return makePlan(start, goal, 0.0, plan);
}

bool OpendriveGlobalPlanner::makePlan(const geometry_msgs::PoseStamped& start, const geometry_msgs::PoseStamped& goal,double tolerance, std::vector<geometry_msgs::PoseStamped>& plan) 
{
  double yaw,best_cost;
  unsigned int best_path_index;
  std::vector<double> x,y,heading;
  std::vector< std::vector<TLaneID> > paths;
  std::vector<double> costs;
  std::vector<TMapPose> start_candidates,end_candidates;
  geometry_msgs::PoseStamped start_out,goal_out,point;
  std::string target_frame;
  std::string source_frame;
  ros::Time time;
  ros::Duration timeout;
  geometry_msgs::TransformStamped transform;

  boost::mutex::scoped_lock lock(mutex_);
  if (!initialized_) {
    ROS_ERROR("OpendriveGlobalPlanner: this planner has not been initialized yet, but it is being used, please call initialize() before use");
    return false;
  }

  //clear the plan, just in case
  plan.clear();

  ros::NodeHandle n;
  std::string global_frame = frame_id_;

  //until tf can handle transforming things that are way in the past... we'll require the goal to be in our global frame
  if (goal.header.frame_id != global_frame) 
  {
    ROS_ERROR("OpendriveGlobalPlanner: the goal pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", global_frame.c_str(), goal.header.frame_id.c_str());
    return false;
  }

  if (start.header.frame_id != global_frame) 
  {
    ROS_ERROR("OpendriveGlobalPlanner: the start pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", global_frame.c_str(), start.header.frame_id.c_str());
    return false;
  }

  /* transform start and end positions to map opendrive */
  try{
    target_frame = this->opendrive_frame_id_;
    source_frame = start.header.frame_id;
    time = start.header.stamp;
    timeout = ros::Duration(0.1);
    if(this->tf2_buffer.canTransform(target_frame, source_frame, time, timeout))
    {
      transform = this->tf2_buffer.lookupTransform(target_frame, source_frame, time);
      tf2::doTransform(start,start_out, transform);
    }else{
      ROS_WARN("OpendriveGlobalPlanner: no transform found for start point from '%s' to '%s'", source_frame.c_str(), target_frame.c_str());
    }
    source_frame = goal.header.frame_id;
    if(this->tf2_buffer.canTransform(target_frame, source_frame, time, timeout))
    {
      transform = this->tf2_buffer.lookupTransform(target_frame, source_frame, time);
      tf2::doTransform(goal,goal_out, transform);
    }else{
      ROS_WARN("OpendriveGlobalPlanner: no transform found for end point from '%s' to '%s'", source_frame.c_str(), target_frame.c_str());
    }
  }catch (tf2::TransformException &ex){
    ROS_ERROR("TF2 Exception: %s",ex.what()); 
  }

  try
  {
    //ROS_WARN("Make Plan");
    yaw=tf2::getYaw(start_out.pose.orientation);
    this->roadmap.set_start_point(start_out.pose.position.x,start_out.pose.position.y,yaw,this->start_dist_tol,this->start_angle_tol);
    this->roadmap.get_start_point_candidates(start_candidates);
    if(this->config.debug)
    {
      std::cout << "--- Start pose candidates: " << std::endl;
      for(unsigned int i=0;i<start_candidates.size();i++)
      {
        std::cout <<"--- " << i << ". x: " << start_candidates[i].point.x << ", y:" << start_candidates[i].point.y << ", heading: " << start_candidates[i].point.heading << std::endl;
        std::cout << "---    distance from desired pose: " << start_candidates[i].distance << std::endl;
        std::cout << "---    lateral offset from desired pose: " << start_candidates[i].lateral_offset << std::endl;
      }
    }
    yaw=tf2::getYaw(goal_out.pose.orientation);
    this->roadmap.set_end_point(goal_out.pose.position.x,goal_out.pose.position.y,yaw,this->end_dist_tol,this->end_angle_tol);
    this->roadmap.get_end_point_candidates(end_candidates);
    if(this->config.debug)
    {
      std::cout << "--- End pose candidates: " << std::endl;
      for(unsigned int i=0;i<end_candidates.size();i++)
      {
        std::cout <<"--- "<< i << ". x: " << end_candidates[i].point.x << ", y:" << end_candidates[i].point.y << ", heading: " << end_candidates[i].point.heading << std::endl;
        std::cout << "---    distance from desired pose: " << end_candidates[i].distance << std::endl;
        std::cout << "---    lateral offset from desired pose: " << end_candidates[i].lateral_offset << std::endl;
      }
    }
    this->roadmap.find_shortest_path(this->multi_hyp);
    this->roadmap.get_paths(costs,paths);
    if(this->config.debug)
    {
      for(unsigned int i=0;i<paths.size();i++)
      {
        std::cout << "--- path " << i << ":" << std::endl;
        for(unsigned int j=0;j<paths[i].size();j++)
          std::cout <<"--- (" << paths[i][j].segment_id << "," << paths[i][j].lane_id << "),";
        std::cout << std::endl;
      }
    }
    best_path_index=this->roadmap.get_best_path(best_cost,this->best_path);
    if(best_path_index==-1)
      return false;
    if(this->config.debug)
    {
      std::cout << "--- best path index: " << best_path_index << std::endl;
      for(unsigned int i=0;i<this->best_path.size();i++)
        std::cout <<"--- (" << this->best_path[i].segment_id << "," << this->best_path[i].lane_id << "),";
      std::cout << std::endl;
    }
    this->best_path_stamp=ros::Time::now();
    this->roadmap.get_trajectory(best_path_index,x,y,heading);
    plan.resize(x.size());
    try{
      target_frame = this->frame_id_;
      source_frame = this->opendrive_frame_id_;
      time = ros::Time(0);
      timeout = ros::Duration(0.1);
      if(this->tf2_buffer.canTransform(target_frame, source_frame, time, timeout))
      {
        transform = this->tf2_buffer.lookupTransform(target_frame, source_frame, time);
        point.header.frame_id=this->frame_id_;
        point.header.stamp=time;
        for(unsigned int i=0;i<x.size();i++)
        {
          point.pose.position.x=x[i];
          point.pose.position.y=y[i];
          point.pose.position.z=0.0;
          tf2::Quaternion orientation;
          orientation.setRPY(0.0,0.0,heading[i]);
          point.pose.orientation=tf2::toMsg(orientation);
          tf2::doTransform(point,plan[i],transform);
          plan[i].header.frame_id=this->frame_id_;
          plan[i].header.stamp=time;
        }
      }else{
        ROS_WARN("OpendriveGlobalPlanner: no transform found for path points from '%s' to '%s'", source_frame.c_str(), target_frame.c_str());
      }
    }catch (tf2::TransformException &ex){
      ROS_ERROR("OpendriveGlobalPlanner: TF2 Exception: %s",ex.what()); 
    }

  }catch(CException &e){
    ROS_ERROR_STREAM(e.what());
    return false;
  }
  //make sure to resize the underlying array that Navfn uses

  //publish the plan for visualization purposes
  publishPlan(plan);
  return !plan.empty();
}

void OpendriveGlobalPlanner::publishPlan(const std::vector<geometry_msgs::PoseStamped>& path) 
{
  if (!initialized_) 
  {
    ROS_ERROR("OpendriveGlobalPlanner: this planner has not been initialized yet, but it is being used, please call initialize() before use");
    return;
  }

  //create a message for the plan
  nav_msgs::Path gui_path;
  gui_path.poses.resize(path.size());

  gui_path.header.frame_id = frame_id_;
  gui_path.header.stamp = ros::Time::now();

  // Extract the plan in world co-ordinates, we assume the path is all in the same frame
  for (unsigned int i = 0; i < path.size(); i++) 
  {
    gui_path.poses[i] = path[i];
  }

  plan_pub_.publish(gui_path);
}

bool OpendriveGlobalPlanner::load_osm_paths(std::vector<TOSMPathData> &osm_paths)
{
  std::vector<std::string> categories;
  std::string name_space;

  /* get the list of categories */
  if(!this->private_nh.getParam("highway_categories", categories))
    return false;
  osm_paths.resize(categories.size());
  for(unsigned int i=0;i<categories.size();i++)
  {
    osm_paths[i].name=categories[i];
    name_space=categories[i]+"/highways";
    if(!this->private_nh.getParam(name_space,osm_paths[i].highways))
      return false;
    name_space=categories[i]+"/keys";
    if(!this->private_nh.getParam(name_space,osm_paths[i].keys))
      return false;
    name_space=categories[i]+"/min_turn_radius";
    if(!this->private_nh.getParam(name_space,osm_paths[i].min_turn_radius))
      osm_paths[i].min_turn_radius=5.0;
    name_space=categories[i]+"/min_road_length";
    if(!this->private_nh.getParam(name_space,osm_paths[i].min_road_length))
      osm_paths[i].min_road_length=0.1;
    name_space=categories[i]+"/use_default_lane_width";
    if(!this->private_nh.getParam(name_space,osm_paths[i].use_default_lane_width))
      osm_paths[i].use_default_lane_width=true;
    name_space=categories[i]+"/default_lane_width";
    if(!this->private_nh.getParam(name_space,osm_paths[i].default_lane_width))
      osm_paths[i].default_lane_width=4.0;
    name_space=categories[i]+"/force_twoway";
    if(!this->private_nh.getParam(name_space,osm_paths[i].force_two_ways))
      return false;
  }

  return true;
}

} //end namespace iri_opendrive_global_planner
